import numpy as np
import cv2

def singleScaleRetinex(img, sigma):
    retinex = np.log10(img) - np.log10(cv2.GaussianBlur(img, (0, 0), sigma))
    return retinex

def multiScaleRetinex(img, sigma_list):
    retinex = np.zeros_like(img)
    for sigma in sigma_list:
        retinex += singleScaleRetinex(img, sigma)
    retinex = retinex / len(sigma_list)
    return retinex

def colorRestoration(img, alpha, beta):
    img_sum = np.sum(img, axis=2, keepdims=True)
    color_restoration = beta * (np.log10(alpha * img) - np.log10(img_sum))
    return color_restoration

def simplestColorBalance(img, low_clip, high_clip):    
    total = img.shape[0] * img.shape[1]
    for i in range(img.shape[2]):
        unique, counts = np.unique(img[:, :, i], return_counts=True)
        current = 0
        for u, c in zip(unique, counts):            
            if float(current) / total < low_clip:
                low_val = u
            if float(current) / total < high_clip:
                high_val = u
            current += c
        img[:, :, i] = np.maximum(np.minimum(img[:, :, i], high_val), low_val)
    return img    

def MSRCR(img, sigma_list, G, b, alpha, beta, low_clip, high_clip):
    img = np.float64(img) + 1.0
    img_retinex = multiScaleRetinex(img, sigma_list)
    img_color = colorRestoration(img, alpha, beta)    
    img_msrcr = G * (img_retinex * img_color + b)
    for i in range(img_msrcr.shape[2]):
        minimum = np.min(img_msrcr[:, :, i])
        maximum = np.max(img_msrcr[:, :, i])
        img_msrcr[:, :, i] = 255 * (img_msrcr[:, :, i] - minimum) / (maximum - minimum)
    img_msrcr = np.uint8(np.minimum(np.maximum(img_msrcr, 0), 255))
    img_msrcr = simplestColorBalance(img_msrcr, low_clip, high_clip)       
    return img_msrcr

if __name__ == '__main__':

    img = cv2.imread("../DinningRoom.jpg",cv2.IMREAD_COLOR)
    
    sigma_list = [15, 80, 250]
    G = 5.0
    b = 25.0
    alpha = 125.0
    beta = 46.0
    low_clip = 0.01
    high_clip = 0.99
    img_msrcr = MSRCR(img,sigma_list,G,b,alpha,beta,low_clip,high_clip)
    
    cv2.imwrite("intrinsic-DinningRoom.jpg",img_msrcr)
    