#include <opencv2/opencv.hpp>

namespace {

    void GaussianBlurViaFFT(const cv::Mat &src, cv::Mat &dst, const double sigma)
    {
        int ksize = cvRound(sigma * 3 * 2 + 1) | 1; // sigma to radius
        cv::Mat kernel1D = cv::getGaussianKernel(ksize, sigma, CV_32F), kernel1D_T;
        cv::transpose(kernel1D, kernel1D_T);
        cv::Mat kernel2D = kernel1D * kernel1D_T;
        cv::Mat _src, _dst;
        src.convertTo(_src, CV_32F);
        cv::filter2D(_src, _dst, CV_32F, kernel2D, cv::Point(-1, -1), 0.0, cv::BORDER_REPLICATE);
        _dst.convertTo(dst, src.depth());
    }

    void singleScaleRetinex(const cv::Mat &src, cv::Mat &dst, double sigma, double a)
    {
        cv::Mat src_blured, log_src, log_src_blured;
        GaussianBlurViaFFT(src, src_blured, sigma);
        cv::log(src, log_src);
        log_src /= log(a);
        cv::log(src_blured, log_src_blured);
        log_src_blured /= log(a);
        cv::subtract(log_src, log_src_blured, dst);
    }

    void MSRCR(const cv::Mat &src, cv::Mat &dst, double sigma, double low_perc, double high_perc, double N, double a)
    {
        cv::Mat img;
        src.convertTo(img, CV_64F);
        img += 1.0;
        cv::Mat img_retinex;
        singleScaleRetinex(img, img_retinex, sigma, a);
        cv::Mat img_levels;
        img_retinex.convertTo(img_levels, CV_32SC1, N);
        std::vector<int> levels(img_levels.total());
        levels.assign(img_levels.data, img_levels.data + img_levels.total());
        std::sort(levels.begin(), levels.end());
        int zero_count = 0;
        std::vector<int> counts(levels.size(), 0);
        for (size_t i = 0, j = 0; i < levels.size(); i++) {
            if (levels[i] == 0) zero_count++;
            counts[j]++;
            if (i + 1 < levels.size() && levels[i] != levels[i + 1]) j++;
        }
        auto last = std::unique(levels.begin(), levels.end());
        levels.erase(last, levels.end());
        counts.resize(levels.size());
        double low_val = levels.front() / N;
        double high_val = levels.back() / N;
        for (size_t j = 0; j < levels.size(); j++) {
            int u = levels[j], c = counts[j];
            if (u < 0 && c < zero_count*low_perc) {
                low_val = u / N;
            }
            if (u > 0 && c < zero_count*high_perc) {
                high_val = u / N;
                break;
            }
        }
        for (auto pix = img_retinex.begin<double>(); pix != img_retinex.end<double>(); pix++) {
            if (*pix < low_val) *pix = low_val;
            if (*pix > high_val) *pix = high_val;
        }
        double minValue, maxValue;
        cv::minMaxLoc(img_retinex, &minValue, &maxValue);
        img_retinex.convertTo(dst, CV_8U, 255.0 / (maxValue - minValue), -minValue * 255.0 / (maxValue - minValue));
    }

}

int main2() {
    cv::Mat img = cv::imread("../DinningRoom.jpg", cv::IMREAD_GRAYSCALE), img_msrcr;

    double sigma = 29;
    double low_clip = 0.01;
    double high_clip = 0.99;
    int N = 160;
    double a = 10.0;
    MSRCR(img, img_msrcr, sigma, low_clip, high_clip, N, a);

    cv::imwrite("comics-DinningRoom.jpg", img_msrcr);
    return 0;
}
